// SPDX-License-Identifier: GPL-2.0+
/* Copyright (c) 2018-2019 AVM GmbH <fritzbox_info@avm.de> */

#include <linux/watchdog.h>
#include <linux/module.h>
#include <avm/watchdog/watchdog.h>
#include <linux/kdebug.h>
#include <linux/sched.h>

#include <asm/avm_enh/avm_enh.h>
#include <avm/enh/reboot_status.h>
#include "avm_sammel.h"

#define DRV_NAME "ar7wdt_hw_shim"

enum wdt_target {
	AR7WDT_TARGET_READY = 0,
	AR7WDT_TARGET_ACTIVE,
	AR7WDT_TARGET_DISABLED
};

struct ar7wdt_shim_info {
	struct watchdog_device *watchdog;
	enum wdt_target         wdt_target;
	struct notifier_block   wdt_device_notifier;
	struct mutex            lock;
};

atomic_t ar7wdt_hw_trigger_cnt;
EXPORT_SYMBOL(ar7wdt_hw_trigger_cnt);

static struct ar7wdt_shim_info ar7wdt_shim_info = {
	.lock = __MUTEX_INITIALIZER(ar7wdt_shim_info.lock),
};

static struct task_struct *current_task_on(int cpu)
{
	return per_cpu(current_task, cpu);
}

static int nmi_notify(struct notifier_block *self,
		      unsigned long cpu, void *param)
{
	struct pt_regs *regs = param;
	struct task_struct *task;
	int init_text = 0;

	task = current_task_on(cpu);
	avm_set_reset_status(RS_NMIWATCHDOG);
	bust_spinlocks(1);
	console_verbose();
	avm_stack_check(NULL);

	printk_avm_console_bend(0); /* force serial-output */

	oops_enter();

	console_verbose();
	bust_spinlocks(1);

	avm_stack_check(task);
	show_task_regs(task, regs);
	add_taint(TAINT_DIE, LOCKDEP_NOW_UNRELIABLE);

	for_each_online_cpu(cpu) {
		regs = nmi_exception_get_cpustacks(cpu);
		if (!regs)
			continue;

		if (init_text == 0) {
			init_text = 1;
			pr_emerg("All other CPU's:\n");
		}
		task = current_task_on(cpu);
		avm_stack_check(task);
		show_task_regs(task, regs);
	}

	avm_oom_show_memstat(0x1);
	/*--- oops_exit triggers kmsg_dump() -> last chance to get infos for panic-log ---*/
	oops_exit();
	panic("Fatal exception: NMI");

	return NOTIFY_STOP;
}

static struct notifier_block nmi_nb = {
	.notifier_call = nmi_notify,
	.priority = INT_MAX
};

static struct ar7wdt_shim_info *ar7wdt_hw_get_shim(void)
{
	return &ar7wdt_shim_info;
}

static int ar7wdt_hw_shim_get_device(struct ar7wdt_shim_info *shim,
				     struct watchdog_device *wdd)
{
	/* Currently this just grabs the first watchdog device seen. */
	if (shim->watchdog) {
		pr_info("%s: Ignoring extra watchdog device (%s)\n", DRV_NAME,
			wdd->info->identity);
		return -EEXIST;
	}

	BUG_ON(!wdd->ops->avm_hw_ping);

	if (!try_module_get(wdd->ops->owner)) {
		pr_err("%s: Can't get watchdog module %s\n", DRV_NAME,
		       wdd->ops->owner->name);
		return -ENODEV;
	}

	wmb();
	shim->watchdog = wdd;
	pr_info("%s: Device attached (%s)\n", DRV_NAME, wdd->info->identity);

	return 0;
}

static int ar7wdt_hw_shim_start_watchdog(struct ar7wdt_shim_info *shim)
{
	int ret;

	pr_info("%s: Starting watchdog\n", DRV_NAME);
	ret = watchdog_start(shim->watchdog);
	if (ret)
		pr_err("%s: Starting watchdog failed: %d\n", DRV_NAME, ret);
	return ret;
}

static int ar7wdt_hw_shim_stop_watchdog(struct ar7wdt_shim_info *shim)
{
	int ret;

	if (!shim->watchdog)
		return 0;

	pr_info("%s: Stopping watchdog\n", DRV_NAME);
	ret = watchdog_stop(shim->watchdog);
	if (ret)
		pr_err("%s: Error stopping watchdog: %d\n", DRV_NAME, ret);

	return ret;
}

static int ar7wdt_hw_shim_put_device(struct ar7wdt_shim_info *shim)
{
	struct watchdog_device *wdd;

	wdd = shim->watchdog;
	if (!wdd)
		return 0;

	shim->watchdog = NULL;
	wmb();
	module_put(wdd->ops->owner);

	return 0;
}

static int wdt_device_notify(struct notifier_block *nb, unsigned long action,
			     void *data)
{
	struct ar7wdt_shim_info *shim =
		container_of(nb, struct ar7wdt_shim_info, wdt_device_notifier);
	struct watchdog_device *wdd = data;
	int notified = NOTIFY_OK;
	int ret;

	mutex_lock(&shim->lock);
	switch (action) {
	case WATCHDOG_NOTIFY_ADD_DEVICE:
		ret = ar7wdt_hw_shim_get_device(shim, wdd);
		if (!ret) {
			if (shim->wdt_target == AR7WDT_TARGET_ACTIVE)
				ar7wdt_hw_shim_start_watchdog(shim);
			else
				ar7wdt_hw_shim_stop_watchdog(shim);
		}
		break;
	case WATCHDOG_NOTIFY_DEL_DEVICE:
		WARN(1, "%s: Unsolicited removal of watchdog device!\n",
		     DRV_NAME);
		ar7wdt_hw_shim_stop_watchdog(shim);
		ar7wdt_hw_shim_put_device(shim);
		break;
	default:
		notified = NOTIFY_DONE;
		break;
	}
	mutex_unlock(&shim->lock);

	return notified;
}

static int __init ar7wdt_hw_shim_init(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();

	shim->wdt_device_notifier.notifier_call = wdt_device_notify;
	watchdog_register_notifier(&shim->wdt_device_notifier);

	return 0;
}

/***** ar7wdt interface functions *****/

void ar7wdt_hw_init(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();

	mutex_lock(&shim->lock);

	if (shim->wdt_target == AR7WDT_TARGET_DISABLED)
		goto out;
	shim->wdt_target = AR7WDT_TARGET_ACTIVE;

	if (shim->watchdog)
		ar7wdt_hw_shim_start_watchdog(shim);
	else
		pr_info("%s: %s: No watchdog device available (yet)\n",
			DRV_NAME, __func__);
out:
	mutex_unlock(&shim->lock);

	register_nmi_notifier(&nmi_nb);
}

int ar7wdt_hw_is_wdt_running(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();
	bool active;

	mutex_lock(&shim->lock);

	if (shim->watchdog)
		active = watchdog_active(shim->watchdog);
	else
		active = false;

	mutex_unlock(&shim->lock);

	return active;
}

void ar7wdt_hw_secure_wdt_disable(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();

	mutex_lock(&shim->lock);

	ar7wdt_hw_shim_stop_watchdog(shim);
	shim->wdt_target = AR7WDT_TARGET_DISABLED;

	mutex_unlock(&shim->lock);
}

void ar7wdt_hw_deinit(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();

	mutex_lock(&shim->lock);

	ar7wdt_hw_shim_stop_watchdog(shim);
	ar7wdt_hw_shim_put_device(shim);

	mutex_unlock(&shim->lock);
}

void ar7wdt_hw_reboot(void)
{
	panic("%s: soft reboot\n", __func__);
}

/* Handle LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0) */
#ifndef smp_mb__before_atomic
#  define smp_mb__before_atomic smp_mb__before_atomic_inc
#endif

void ar7wdt_hw_trigger(void)
{
	struct ar7wdt_shim_info *shim = ar7wdt_hw_get_shim();
	struct watchdog_device *wdd = shim->watchdog;

	/* This must not sleep or take locks */
	if (wdd && wdd->ops->avm_hw_ping)
		wdd->ops->avm_hw_ping(wdd);

	smp_mb__before_atomic();
	atomic_inc(&ar7wdt_hw_trigger_cnt);
}

module_init(ar7wdt_hw_shim_init);
