// SPDX-License-Identifier: GPL-2.0
#define pr_fmt(fmt) "[avm_prom_config] " fmt

#pragma GCC push_options
#pragma GCC diagnostic ignored "-Wunused-parameter"
#pragma GCC diagnostic ignored "-Wsign-compare"
#include <linux/module.h>

#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <linux/genhd.h>
#include <linux/zlib.h>
#include <linux/of.h>
#include <linux/pagemap.h>
#include <linux/blkdev.h>
#pragma GCC pop_options

#include <avm/enh/prom_config.h>
#include <avm/enh/prom_config_legacy.h>
#include "prom_config_internal.h"

struct ondisk_config_hdr {
	u8 version;
	u8 type;
	__be16 len;
} __packed;

struct config_data {
	loff_t offset;
	size_t size;
	enum avm_prom_config_type zipped_src_type;

	// Flags
	unsigned int is_present : 1;
	unsigned int has_zipped_src : 1;
};

// This struct holds information about the zip types
//
// It will be filled by load_config. If a non-zipped config
// is found on-disk the has_zipped_src flag is cleared.
static struct config_data configs[AVM_PROM_NUM_TYPES] = {
	[WLAN] = { .has_zipped_src = 1, .zipped_src_type = WLAN_ZIP },
	[WLAN2] = { .has_zipped_src = 1, .zipped_src_type = WLAN2_ZIP },
	[WLAN3] = { .has_zipped_src = 1, .zipped_src_type = WLAN3_ZIP },
};

struct avm_prom_backend {
	union {
		struct mtd_info *mtd;
		struct block_device *bd;
	} config;
	int (*read)(loff_t offset, void *buf, size_t size);
	loff_t (*lba_to_ofs)(u32 lba);
};

static struct avm_prom_backend avm_prom_backend;

int avm_prom_is_config_available(enum avm_prom_config_type type)
{
	struct config_data *config;

	if (type >= AVM_PROM_NUM_TYPES)
		return 0;

	config = &configs[type];

	if (config->has_zipped_src)
		return avm_prom_is_config_available(config->zipped_src_type);

	return config->is_present;
}

int get_wlan_dect_config(enum wlan_dect_type type, unsigned char *buffer,
			 unsigned int len)
{
	struct wlan_dect_config *hdr = (void *)buffer;
	ssize_t ret;

	ret = avm_prom_get_config(type, hdr + 1,
				  len - sizeof(struct wlan_dect_config));

	if (ret < 0)
		return ret;

	// We can not return the size, so abort
	if (ret >= U16_MAX)
		return -EOVERFLOW;

	// Fake the header
	hdr->Version = 1;
	hdr->Type = type;
	hdr->Len = ret;

	return 0;
}
EXPORT_SYMBOL(get_wlan_dect_config);

static int config_mtd_read(loff_t offset, void *buf, size_t size)
{
	size_t retlen;
	int ret;
	u8 align_buf[4];

	// The start is unaligned, so we align our reads to 4
	if (!IS_ALIGNED(offset, sizeof(align_buf))) {
		loff_t aligned_offset = round_down(offset, sizeof(align_buf));
		unsigned int start = offset - aligned_offset;
		unsigned int len = min(sizeof(align_buf) - start, size);

		ret = mtd_read(avm_prom_backend.config.mtd, aligned_offset,
				sizeof(align_buf), &retlen, align_buf);

		if (ret < 0)
			return ret;

		// We need to copy over the last len bytes
		memcpy(buf, align_buf + start, len);
		offset = ALIGN(offset, sizeof(align_buf));
		buf += len;
		size -= len;
	}

	// Now copy aligned to 4 byte boundaries
	if (size >= sizeof(align_buf)) {
		size_t len = round_down(size, sizeof(align_buf));

		ret = mtd_read(avm_prom_backend.config.mtd, offset, len,
				&retlen, buf);

		if (ret < 0)
			return ret;

		offset += len;
		buf += len;
		size -= len;
	}

	// Read the last remaining bytes
	if (size > 0) {
		ret = mtd_read(avm_prom_backend.config.mtd, offset,
				sizeof(align_buf), &retlen, align_buf);

		if (ret < 0)
			return ret;

		memcpy(buf, align_buf, size);
	}

	return 0;
}

static loff_t config_mtd_lba_to_ofs(u32 lba)
{
	return lba;
}

/**
 * Explicitly set the config mtd.
 *
 * @param mtd MTD to consult for config
 */
static int avm_prom_set_config_mtd(struct mtd_info *mtd)
{
	avm_prom_backend.config.mtd = mtd;
	avm_prom_backend.lba_to_ofs = config_mtd_lba_to_ofs;
	wmb();  // Make sure "read" pointer gets written last so backend
	// is completely initialized before any of this is used in avm_prom_read
	avm_prom_backend.read = config_mtd_read;

	pr_debug("Use config mtd %s (mtd%d)\n", mtd->name,
	       mtd->index);

	return 0;
}

static int config_blk_read(loff_t offset, void *buf, size_t size)
{
	pgoff_t index;
	off_t page_offset;

	index = offset >> PAGE_SHIFT;
	page_offset = offset & (PAGE_SIZE - 1);

	while (size > 0) {
		size_t cpylen;
		struct page *page;

		if ((page_offset + size) > PAGE_SIZE)
			cpylen = PAGE_SIZE - page_offset;
		else
			cpylen = size;

		page = read_mapping_page(
				avm_prom_backend.config.bd->bd_inode->i_mapping,
				index,
				NULL);
		if (IS_ERR(page))
			return PTR_ERR(page);

		memcpy(buf, page_address(page) + page_offset, cpylen);

		put_page(page);

		buf += cpylen;
		size -= cpylen;
		page_offset = 0;
		index++;
	}

	return 0;
}

static loff_t config_blk_lba_to_ofs(u32 lba)
{
	return lba * bdev_physical_block_size(avm_prom_backend.config.bd);
}

/**
 * Explicitly set the config block device.
 *
 * @param blk Block device to consult for config
 */
static int avm_prom_set_config_blk(struct block_device *blk)
{
	char b[BDEVNAME_SIZE];

	avm_prom_backend.config.bd = blk;
	avm_prom_backend.lba_to_ofs = config_blk_lba_to_ofs;
	wmb();  // Make sure "read" pointer gets written last so backend
	// is completely initialized before any of this is used in avm_prom_read
	avm_prom_backend.read = config_blk_read;

	pr_debug("Use config blkdev %s (%s)\n",
	       blk->bd_part->info->volname, bdevname(blk, b));

	return 0;
}

static const char *avm_prom_get_config_device_name(void)
{
	const char *name;
	int ret;

	if (!of_have_populated_dt() || !of_chosen)
		goto out_default;

	ret = of_property_read_string(of_chosen, "wlan_dect_config_part",
				      &name);
	if (ret < 0)
		goto out_default;

	return name;

out_default:
	return "urlader";
}

static int match_dev_by_label(struct device *dev, const void *data)
{
	const char *label = data;
	struct hd_struct *part = dev_to_part(dev);

	if (!part->info)
		return 0;

	if (strcasecmp(label, part->info->volname))
		return 0;

	// TODO: we should make sure that we have an internal (non-usb) device
	return 1;
}

static struct block_device *blkdev_get_by_name(const char *name)
{
	struct device *dev;

	wait_for_device_probe();

	dev = class_find_device(&block_class, NULL, name, &match_dev_by_label);

	if (!dev)
		return ERR_PTR(ENODEV);

	return blkdev_get_by_dev(dev->devt, FMODE_READ, THIS_MODULE);
}

static int avm_prom_detect_config_device(void)
{
	const char *name;
	struct mtd_info *mtd;
	struct block_device *blk;

	name = avm_prom_get_config_device_name();

	// First try if a mtd exists
	mtd = get_mtd_device_nm(name);

	// Exist, so use it
	if (!IS_ERR(mtd)) {
		avm_prom_set_config_mtd(mtd);

		return 0;
	}

	blk = blkdev_get_by_name(name);

	if (!IS_ERR(blk)) {
		avm_prom_set_config_blk(blk);

		return 0;
	}

	pr_err("Could not find config device\n");
	return -ENODEV;
}

static int avm_prom_read(loff_t offset, void *buf, size_t size)
{
	if (!avm_prom_backend.read)
		return -ENODEV;
	rmb();  // Make sure backend is tested for completed initialization
	// before eny other part is accessed below.

	return avm_prom_backend.read(offset, buf, size);
}

static ssize_t avm_prom_load_config_entry_internal(loff_t offset)
{
	int ret, i;
	struct ondisk_config_hdr hdr;
	ssize_t len = 0;
	struct config_data *config;

	BUILD_BUG_ON(sizeof(struct ondisk_config_hdr) != 4);
	ret = avm_prom_read(offset, &hdr, sizeof(hdr));
	if (ret < 0)
		return ret;

	switch (hdr.version) {
	case 1: // FALL THROUGH
	case 2:
		len = be16_to_cpu(hdr.len);
		break;
	case 0xFF:
		return -ENOENT;
	default:
		return -EINVAL;
	}

	if (hdr.type >= AVM_PROM_NUM_TYPES)
		return -EINVAL;

	config = &configs[hdr.type];

	// We already loaded a config with this type
	if (config->is_present)
		return -EBUSY;

	config->offset = offset + sizeof(hdr);
	config->size = len;
	config->is_present = 1;
	config->has_zipped_src = 0;

	pr_debug("Add entry ofs=0x%llx type=%d size=0x%x\n",
	       config->offset, hdr.type, config->size);

	avm_prom_create_config_proc(hdr.type);

	for (i = 0; i < AVM_PROM_NUM_TYPES; i++) {
		if (configs[i].has_zipped_src && configs[i].zipped_src_type == hdr.type)
			avm_prom_create_config_proc(i);
	}

	return config->size + sizeof(hdr);
}

int avm_prom_load_config_entry(loff_t offset)
{
	ssize_t ret;

	if (!avm_prom_backend.read) {
		ret = avm_prom_detect_config_device();

		if (ret < 0)
			return ret;
	}

	ret = avm_prom_load_config_entry_internal(offset);

	// Conceal returned size
	return min(ret, 0);
}

int avm_prom_load_config_entries(loff_t offset, size_t len)
{
	ssize_t ret;

	if (!avm_prom_backend.read) {
		ret = avm_prom_detect_config_device();

		if (ret < 0)
			return ret;
	}

	while (len > 0) {
		ret = avm_prom_load_config_entry_internal(offset);

		// Nothing more to read
		if (ret == -ENOENT) {
			break;
		} else if (ret < 0) {
			pr_err("Error reading entry: %d\n", ret);
			return ret;
		}

		// Advance our starting point and try next entry
		len -= ret;
		offset += ret;
	}

	return 0;
}

/**
 * Automatically read config offsets from the devicetree
 */
int avm_prom_load_config_devicetree(void)
{
	int count, i, ret;
	const __be32 *lbas;

	if (!IS_ENABLED(CONFIG_OF_AVM_DT_ENV))
		return 0;

	if (!of_have_populated_dt() || !of_chosen) {
		pr_err("Don't have chosen node in DT\n");
		return -ENODEV;
	}

	if (!avm_prom_backend.read) {
		ret = avm_prom_detect_config_device();

		if (ret)
			return ret;
	}

	lbas = of_get_property(of_chosen, "wlan_dect_configs", &count);
	if (!lbas) {
		pr_err("No wlan_dect_configs property in chosen node\n");
		return -EINVAL;
	}

	count /= sizeof(__be32);

	for (i = 0; i < count; i++) {
		ret = avm_prom_load_config_entry(
			avm_prom_backend.lba_to_ofs(be32_to_cpu(lbas[i])));

		if (ret != 0 && ret != -ENOENT) {
			pr_err("Error reading entry %d: %d\n", i, ret);
			return ret;
		}
	}

	return 0;
}

#if defined(CONFIG_OF_AVM_DT_ENV)
late_initcall(avm_prom_load_config_devicetree);
#endif /*--- defined(CONFIG_OF_AVM_DT_ENV) ---*/

static ssize_t avm_prom_get_config_raw_alloc(enum avm_prom_config_type type,
					     void **buf, size_t max_len)
{
	bool alloced_buffer = false;
	struct config_data *config;
	ssize_t ret;
	size_t len;

	if (type >= AVM_PROM_NUM_TYPES)
		return -EINVAL;

	config = &configs[type];

	// We don't handle zip here
	if (config->has_zipped_src)
		return -EINVAL;

	if (!config->is_present)
		return -ENOENT;

	if (!*buf) {
		*buf = vmalloc(config->size);

		if (!*buf)
			return -ENOMEM;

		alloced_buffer = true;
		max_len = config->size;
	}

	len = min(config->size, max_len);

	ret = avm_prom_read(config->offset, *buf, len);
	if (ret < 0) {
		if (alloced_buffer)
			vfree(*buf);

		return ret;
	}

	// Zero out remaining space
	if (max_len > len)
		memset(*buf + len, 0, max_len - len);

	return config->size;
}

static ssize_t get_inflated_size(void *zipped_buf, size_t zipped_size,
		z_stream *stream)
{
	ssize_t ret;
	char tbuf[32];

	stream->next_in = zipped_buf;
	stream->avail_in = zipped_size;

	stream->next_out = tbuf;
	stream->avail_out = sizeof(tbuf);

	// Simulate decompression
	while ((ret = zlib_inflate(stream, Z_SYNC_FLUSH)) == Z_OK) {
		stream->next_out = tbuf;
		stream->avail_out = sizeof(tbuf);
	}

	if (ret != Z_STREAM_END)
		return -EIO;

	ret = stream->total_out;
	zlib_inflateReset(stream);

	return ret;
}

static ssize_t unzip_buffer(void **buf, size_t max_len,
		void *zipped_buf, size_t zipped_size)
{
	ssize_t ret;
	bool alloced_buffer = false;
	z_stream stream;

	stream.workspace = vmalloc(zlib_inflate_workspacesize());

	if (!stream.workspace)
		return -ENOMEM;

	ret = zlib_inflateInit(&stream);
	if (ret != Z_OK) {
		ret = -EIO;
		goto out_err_zlib_init;
	}

	// No buffer supplied, so guess the size
	if (!*buf) {
		ret = get_inflated_size(zipped_buf, zipped_size, &stream);

		if (ret < 0)
			goto out_err_zlib;

		// Get out the size and alloc the buffer
		max_len = ret;
		*buf = vmalloc(max_len);
		if (!*buf) {
			ret = -ENOMEM;
			goto out_err_zlib;
		}
		alloced_buffer = true;
	}

	stream.next_in = zipped_buf;
	stream.avail_in = zipped_size;

	stream.next_out = *buf;
	stream.avail_out = max_len;

	ret = zlib_inflate(&stream, Z_FINISH);
	if (ret == Z_OK) {
		zlib_inflateReset(&stream);
		ret = get_inflated_size(zipped_buf, zipped_size, &stream);
		// Either we hold error code or total size in ret here.
		goto out_err_zlib;
	}
	if (ret != Z_STREAM_END) {
		ret = -EIO;
		goto out_err_zlib;
	}

	// Store actual size of output
	ret = stream.total_out;

	// Zero out remainder
	if ((size_t)ret < max_len)
		memset(*buf + ret, 0, max_len - ret);

out_err_zlib:
	// We error-out and have allocated the buffer.
	if (ret < 0 && alloced_buffer) {
		vfree(*buf);
		*buf = NULL;
	}

	zlib_inflateEnd(&stream);

out_err_zlib_init:
	vfree(stream.workspace);

	return ret;
}

static ssize_t avm_prom_get_config_zip_alloc(enum avm_prom_config_type type,
					     void **buf, size_t max_len)
{
	struct config_data *config;
	ssize_t ret, zipped_size;
	void *zipped_buf = NULL;

	if (type >= AVM_PROM_NUM_TYPES)
		return -EINVAL;

	config = &configs[type];

	// Not a zip file, so we can just use the raw read
	if (!config->has_zipped_src)
		return -EINVAL;

	// Read zipped buf into temporary buffer
	zipped_size = avm_prom_get_config_raw_alloc(config->zipped_src_type,
							 &zipped_buf, 0);
	if (zipped_size < 0)
		return zipped_size;

	ret = unzip_buffer(buf, max_len, zipped_buf, zipped_size);
	vfree(zipped_buf);

	return ret;
}

static ssize_t
avm_prom_get_config_internal_alloc(enum avm_prom_config_type type, void **buf,
				   size_t max_len)
{
	struct config_data *config;
	ssize_t ret;

	if (type >= AVM_PROM_NUM_TYPES)
		return -EINVAL;

	config = &configs[type];

	if (config->has_zipped_src)
		ret = avm_prom_get_config_zip_alloc(type, buf, max_len);
	else
		ret = avm_prom_get_config_raw_alloc(type, buf, max_len);

	return ret;
}

ssize_t avm_prom_get_config(enum avm_prom_config_type type, void *buf,
			    size_t max_len)
{
	void *_buf = buf;

	return avm_prom_get_config_internal_alloc(type, &_buf, max_len);
}
EXPORT_SYMBOL(avm_prom_get_config);

ssize_t avm_prom_get_config_alloc(enum avm_prom_config_type type, void **_buf)
{
	*_buf = NULL;

	return avm_prom_get_config_internal_alloc(type, _buf, 0);
}
EXPORT_SYMBOL(avm_prom_get_config_alloc);
