// SPDX-License-Identifier: GPL-2.0+
#define pr_fmt(fmt) "[avm_prom_config] " fmt

#pragma GCC push_options
#pragma GCC diagnostic ignored "-Wsign-compare"
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/slab.h>
#pragma GCC pop_options

#include <avm/enh/prom_config.h>
#include "prom_config_internal.h"

static struct proc_dir_entry *calibprocdir;

#define PROC_CALIBDIR "avm/calib"

// clang-format off
static struct _calib_list {
	enum avm_prom_config_type type;
	const char *proc_device;
	struct proc_dir_entry *entry;
} calib_list[] = {
	{ .type = WLAN,   .proc_device = "wlan"},
	{ .type = WLAN2,  .proc_device = "wlan2"},
	{ .type = WLAN3,  .proc_device = "wlan3"},
	{ .type = DECT,   .proc_device = "dect"},
	{ .type = DOCSIS, .proc_device = "docsis"},
	{ .type = DSL,    .proc_device = "dsl"},
	{ .type = AVMZERTIFIKATE, .proc_device = "avmzertifikate"},
};
// clang-format on

struct private_data {
	void *buffer;
	size_t size;
};

static int avm_prom_config_open(struct inode *inode, struct file *file)
{
	struct private_data *data;
	ssize_t ret;
	struct _calib_list *info = PDE_DATA(inode);

	WARN_ON(file->private_data);

	data = kmalloc(sizeof(struct private_data), GFP_KERNEL);

	if (!data)
		return -ENOMEM;

	ret = avm_prom_get_config_alloc(info->type, &data->buffer);
	if (ret < 0) {
		kfree(data);
		return ret;
	}

	data->size = ret;

	file->private_data = data;

	return 0;
}

static int avm_prom_config_read(struct file *file, char __user *buf, size_t len, loff_t *ofs)
{
	struct private_data *data = file->private_data;
	size_t read_len;

	if (*ofs >= data->size)
		return 0;

	read_len = min_t(size_t, len, data->size - *ofs);

	if (copy_to_user(buf, data->buffer + *ofs, read_len))
		return -EFAULT;

	*ofs += read_len;

	return read_len;
}

static int avm_prom_config_release(struct inode *inode, struct file *file)
{
	struct private_data *data = file->private_data;

	vfree(data->buffer);
	kfree(data);

	return 0;
}

static const struct file_operations fops = {
	.open = avm_prom_config_open,
	.read = avm_prom_config_read,
	.release = avm_prom_config_release,
};

int avm_prom_create_config_proc(enum avm_prom_config_type type)
{
	struct _calib_list *c = NULL;
	unsigned int i;

	if (!calibprocdir)
		calibprocdir = proc_mkdir(PROC_CALIBDIR, NULL);

	if (!calibprocdir) {
		pr_err("%s: can't create %s\n", __func__, PROC_CALIBDIR);
		return -ENOMEM;
	}

	for (i = 0; i < ARRAY_SIZE(calib_list); i++) {
		if (calib_list[i].type == type) {
			c = &calib_list[i];
			break;
		}
	}

	// Don't want an entry, skip
	if (!c)
		return 0;

	// Already have an entry, skip
	if (c->entry)
		return 0;

	c->entry = proc_create_data(c->proc_device, 0, calibprocdir, &fops, c);

	if (!c->entry)
		return -EIO;

	return 0;
}
