/*
 * Driver for EIP97 cryptographic accelerator.
 *
 * Copyright (c) 2018 AVM GmbH <fritzbox_info@avm.de>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This driver is heavily based on the Mediatek driver from Ryder Lee.
 *
 */

#ifndef __EIP97_AEAD_RAW_H_
#define __EIP97_AEAD_RAW_H_

#include "eip97-platform.h"

#define EIP97_AEAD_MAX_CT_SIZE			6

#define EIP97_AEAD_MAX_STATE_BUF_SIZE		16

struct eip97_aead_info {
	u32 cmd[EIP97_AEAD_MAX_CT_SIZE];
	u32 tfm[2];
	u32 state[EIP97_AEAD_MAX_STATE_BUF_SIZE];
};

/**
 * struct eip97_aes_gcm_request - Parameters for eip97_aes_gcm_crypt()
 * @key_words:	Number of words in the en-/decryption key
 * @key:	Pointer to the en-/decryption key
 * @prehash:	Pointer to four words of precomputed hash, which is the
 *		result of AES in CTR mode on all-zero IV, all-zero data
 *		using the current key @key.
 * @iv_bytes:	Number of bytes in the IV
 * @iv:		Pointer to the IV
 * @auth_bytes:	Number of bytes in the authentication tag
 * @assoc_bytes:Number of bytes in the associated data
 * @src:	Pointer to the source scatterlist that already holds DMA
 *		addresses.
 *		For encryption this scatterlist holds:
 *
 *		assoc data | plaintext data
 *
 *		For decryption this scatterlist holds:
 *
 *		assoc data | encrypted data | authentication tag
 *
 * @src_bytes:	Number of bytes referenced by @src.
 * @dst:	Pointer to the destination scatterlist that already
 *		holds DMA addresses.
 *		For encryption this scatterlist is filled with:
 *
 *		assoc data | encrypted data | authentication tag
 *
 *		For decryption this scatterlist is filled with:
 *
 *		assoc data | decrypted data
 *
 * @dst_bytes:	Number of bytes referenced by @dst.
 * @dir:	Direction of crypto operation. Either encryption
 *		(EIP97_ENCRYPT) or decryption (EIP97_DECRYPT).
 * @tag:	Identification tag for the current crypto operation.
 * @base:	Pointer to the register base of the EIP97 crypto device
 *		registers. The yield thread should use
 *		eip97_find_dev()->base.
 * @ring_id:	Number of ring to use. The Linux kernel driver uses ring
 *		EIP97_RING0, the yield thread should use ring
 *		EIP97_RING1.
 * @ring:	Pointer to the ring structure allocated by the kernel.
 *		The yield thread should use
 *		eip97_find_dev()->ring[EIP97_RING1].
 * @info:	Pointer to a context structure used to execute the
 *		request. This should be a virtual address.
 * @info_dma:	DMA address of the context structure.
 * @sync:	Function to be called after construction of the crypto
 *		request, before triggering the EIP97 crypto hardware
 *		for processing. Useful i.e. if @info_dma is a
 *		streaming DMA address and needs to be synced after
 *		@info was constructed. @sync can be set to 0
 *		if no such synchronisation is needed.
 * @sync_ctx:	Context pointer to be passed to @sync.
 *
 **/
struct eip97_aes_gcm_request {
	u32 key_words;
	u32 *key;
	u32 *prehash;
	u32 iv_bytes;
	u8 *iv;
	u32 auth_bytes;
	u32 assoc_bytes;
	struct scatterlist *src;
	u32 src_bytes;
	struct scatterlist *dst;
	u32 dst_bytes;
	enum crypt_direction dir;
	u32 tag;
	void __iomem *base;
	int ring_id;
	struct eip97_ring *ring;
	struct eip97_aead_info *info;
	dma_addr_t info_dma;
	void (*sync)(void *ctx);
	void *sync_ctx;
};

int eip97_aes_gcm_crypt(struct eip97_aes_gcm_request *req);

#endif /* __EIP97_AEAD_RAW_H_ */
