/*
 * Driver for EIP97 cryptographic accelerator.
 *
 * Copyright (c) 2018 AVM GmbH <fritzbox_info@avm.de>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This driver is heavily based on the Mediatek driver from Ryder Lee.
 *
 */

#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <crypto/scatterwalk.h>
#include <linux/scatterlist.h>
#include <crypto/aes.h>
#include <crypto/skcipher.h>
#include <crypto/internal/aead.h>
#include "eip97-platform.h"
#include "eip97-utest.h"
#include "eip97-aead.h"
#include "eip97-aead-raw.h"

#define EIP97_AEAD_MAX_KEY_SIZE			8

#define EIP97_AES_GCM_PREHASH_WORDS		4

#define EIP97_AES_GCM_CT_USE_EIP97_MODE		0x00020000

#define EIP97_AES_GCM_CT_CONTAINS_IV		0x1c000000

struct eip97_aead_tfm_ctx {
	struct eip97_cryp *cryp;

	u32 key_words;
	u32 key[EIP97_AEAD_MAX_KEY_SIZE];
	u32 prehash[EIP97_AES_GCM_PREHASH_WORDS];
	u32 auth_bytes;

	struct crypto_skcipher *aes_ctr;
};

enum aead_mode {
	AES_GCM,
};

struct eip97_aead_sg {
	u32 proc_bytes;
	struct scatterlist *unaligned_sg;
	u32 unaligned_nents;
	u32 nents;
	u32 dma_nents;
};

struct eip97_aead_req_ctx {
	struct eip97_aead_sg src;
	struct eip97_aead_sg dst;
	enum aead_mode mode;
	enum crypt_direction dir;
	struct eip97_aead_info info;
	dma_addr_t info_dma;
	enum eip97_op_type type;
};

struct eip97_aes_gcm_setkey_result {
	int err;
	struct completion completion;
};

struct eip97_aead_info_dma_sync {
	struct device *dev;
	dma_addr_t addr;
};

static inline struct eip97_aead_req_ctx *
req_ctx(struct aead_request *req)
{
	return (struct eip97_aead_req_ctx *) req->__ctx;
}

static void eip97_aes_gcm_setkey_done(struct crypto_async_request *req, int err)
{
	struct eip97_aes_gcm_setkey_result *result = req->data;

	if (err == -EINPROGRESS)
		return;

	result->err = err;
	complete(&result->completion);
}

static int eip97_aes_gcm_setkey(struct crypto_aead *aead,
				const u8 *key, u32 key_bytes)
{
	struct eip97_aead_tfm_ctx *ctx = crypto_tfm_ctx(&aead->base);
	struct crypto_skcipher *skc = ctx->aes_ctr;
	int err;

	struct {
		u32 hash[4];
		u8 iv[8];

		struct eip97_aes_gcm_setkey_result result;

		struct scatterlist sg[1];
		struct skcipher_request req;
	} *data;

	switch (key_bytes) {
	case AES_KEYSIZE_128:
	case AES_KEYSIZE_192:
	case AES_KEYSIZE_256:
		break;
	default:
		crypto_aead_set_flags(aead, CRYPTO_TFM_RES_BAD_KEY_LEN);
		return -EINVAL;
	}

	/* For some mysterious hardware-related reason we need to precompute
	 * some kind of hash from all-zero IV, all-zero data and current key by
	 * using AES in CTR mode.
	 */

	crypto_skcipher_clear_flags(skc, CRYPTO_TFM_REQ_MASK);
	crypto_skcipher_set_flags(skc, crypto_aead_get_flags(aead) &
				  CRYPTO_TFM_REQ_MASK);

	err = crypto_skcipher_setkey(skc, key, key_bytes);
	if (err)
		return err;

	crypto_aead_set_flags(aead, crypto_skcipher_get_flags(skc) &
			      CRYPTO_TFM_RES_MASK);

	data = kzalloc(sizeof(*data) + crypto_skcipher_reqsize(skc),
		       GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	init_completion(&data->result.completion);
	sg_init_one(data->sg, &data->hash, AES_BLOCK_SIZE);

	skcipher_request_set_tfm(&data->req, skc);
	skcipher_request_set_callback(&data->req, CRYPTO_TFM_REQ_MAY_SLEEP |
			CRYPTO_TFM_REQ_MAY_BACKLOG,
			eip97_aes_gcm_setkey_done, &data->result);
	skcipher_request_set_crypt(&data->req, data->sg, data->sg,
			AES_BLOCK_SIZE, data->iv);

	err = crypto_skcipher_encrypt(&data->req);
	if (err == -EINPROGRESS) {
		err = wait_for_completion_interruptible(
				&data->result.completion);
		if (!err)
			err = data->result.err;
	}
	if (err)
		goto exit;

	ctx->key_words = SIZE_IN_WORDS(key_bytes);

	BUILD_BUG_ON(sizeof(ctx->key) < AES_KEYSIZE_256);
	memcpy(ctx->key, key, key_bytes);

	memcpy(ctx->prehash, data->hash, sizeof(ctx->prehash));

exit:
	kzfree(data);
	return err;
}

static int eip97_aes_gcm_setauthsize(struct crypto_aead *aead, u32 auth_bytes)
{
	struct eip97_aead_tfm_ctx *ctx = crypto_tfm_ctx(&aead->base);

	switch (auth_bytes) {
	case 8:
	case 12:
	case 16:
		break;
	default:
		return -EINVAL;
	}

	ctx->auth_bytes = auth_bytes;
	return 0;
}

static int eip97_aead_complete(struct aead_request *req, int err, bool callback)
{
	struct crypto_async_request *areq = &req->base;

	if (callback)
		areq->complete(areq, err);

	return err;
}

static int eip97_aead_dma_unalign(struct aead_request *req, int err,
				  bool callback)
{
	struct eip97_aead_req_ctx *ctx = req_ctx(req);
	size_t copied = 0;
	u8 *align_buf = 0;

	if (req->src != ctx->src.unaligned_sg) {
		align_buf = sg_virt(req->src);

		req->src = ctx->src.unaligned_sg;
		ctx->src.nents = ctx->src.unaligned_nents;
	}

	if (req->dst != ctx->dst.unaligned_sg) {
		align_buf = sg_virt(req->dst);

		req->dst = ctx->dst.unaligned_sg;
		ctx->dst.nents = ctx->dst.unaligned_nents;

		if (err)
			goto exit;

		copied = sg_copy_from_buffer(req->dst, ctx->dst.nents,
					     align_buf, ctx->dst.proc_bytes);
		if (copied != ctx->dst.proc_bytes)
			err = -ENODATA;
	}

exit:
	if (align_buf)
		kfree(align_buf);

	return eip97_aead_complete(req, err, callback);
}

static int eip97_aead_dma_unmap(struct aead_request *req, int err,
				bool callback)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);
	struct device *dev = tfm_ctx->cryp->dev;

	dma_unmap_single(dev, ctx->info_dma, sizeof(ctx->info), DMA_TO_DEVICE);

	dma_unmap_sg(dev, req->src, ctx->src.dma_nents, DMA_TO_DEVICE);

	if (req->src != req->dst)
		dma_unmap_sg(dev, req->dst, ctx->dst.dma_nents, DMA_FROM_DEVICE);

	return eip97_aead_dma_unalign(req, err, callback);
}

void eip97_aead_finish(enum eip97_op_type *op_type, eip97_err_t err,
		       unsigned int bytes)
{
	struct eip97_aead_req_ctx *ctx = container_of(op_type,
					 struct eip97_aead_req_ctx, type);

	struct aead_request *req = container_of((void *)ctx,
				   struct aead_request, __ctx);

	if (err & EIP97_ERR_CHECK_FAIL) {
		eip97_aead_dma_unmap(req, -EBADMSG, true);
		return;
	} else if (err & EIP97_ERR_TIMEOUT) {
		eip97_aead_dma_unmap(req, -ETIMEDOUT, true);
		return;
	} else if (err) {
		eip97_aead_dma_unmap(req, -EPROTO, true);
		return;
	}

	if (bytes != ctx->dst.proc_bytes)
		eip97_aead_dma_unmap(req, -ENODATA, true);
	else
		eip97_aead_dma_unmap(req, 0, true);
}

static int eip97_aes_gcm_create_tfm(struct eip97_aes_gcm_request *req)
{
	u32 iv_words;

	req->info->tfm[0] = 0;
	req->info->tfm[1] = 0;

	switch (SIZE_IN_BYTES(req->key_words)) {
	case AES_KEYSIZE_128:
		req->info->tfm[0] |= EIP97_TFM_AES_128BITS;
		break;
	case AES_KEYSIZE_192:
		req->info->tfm[0] |= EIP97_TFM_AES_192BITS;
		break;
	case AES_KEYSIZE_256:
		req->info->tfm[0] |= EIP97_TFM_AES_256BITS;
		break;
	default:
		return -EINVAL;
	}

	switch (req->iv_bytes) {
	case 16:
		req->info->tfm[1] |= EIP97_TFM_IV3;
	case 12:
		req->info->tfm[1] |= EIP97_TFM_IV2;
	case 8:
		req->info->tfm[1] |= EIP97_TFM_IV1;
	case 4:
		req->info->tfm[1] |= EIP97_TFM_IV0;
		iv_words = req->iv_bytes / 4;
		break;
	default:
		return -EINVAL;
	}

	req->info->tfm[0] |= EIP97_TFM_SIZE(req->key_words
			   + EIP97_AES_GCM_PREHASH_WORDS
			   + iv_words);

	BUILD_BUG_ON(sizeof(req->info->state) < AES_KEYSIZE_256 +
		     SIZE_IN_BYTES(EIP97_AES_GCM_PREHASH_WORDS) + 16);

	eip97_write_state_le(req->info->state, req->key, req->key_words);
	req->info->tfm[0] |= EIP97_TFM_CONTAINS_KEY;

	eip97_write_state_be(req->info->state + req->key_words, req->prehash,
			     EIP97_AES_GCM_PREHASH_WORDS);

	eip97_write_state_le(req->info->state + req->key_words + EIP97_AES_GCM_PREHASH_WORDS,
			     (u32 *)req->iv, iv_words);

	switch (req->dir) {
	case EIP97_ENCRYPT:
		req->info->tfm[0] |= EIP97_TFM_CRYPT_THEN_HASH_OUT;
		break;
	case EIP97_DECRYPT:
		req->info->tfm[0] |= EIP97_TFM_HASH_THEN_CRYPT_IN;
		break;
	default:
		return -EINVAL;
	}

	/* For whatever reason, SHA-224 flag must be set. */
	req->info->tfm[0] |= EIP97_TFM_AES_AEAD_DIGEST | EIP97_TFM_SHA224;

	req->info->tfm[1] |= EIP97_TFM_COUNTER_MODE_IV |
			     EIP97_TFM_ENCRYPT_HASH_RESULT |
			     EIP97_TFM_CRYPT_MODE_UNDEFINED;
	return 0;
}

static u32 eip97_aes_gcm_create_command_token(struct eip97_aes_gcm_request *req,
					      u32 text_bytes)
{
	u32 cmd_words = 0;

	BUILD_BUG_ON(sizeof(req->info->cmd) < SIZE_IN_BYTES(6));

	req->info->cmd[cmd_words++] = 0x0b000000 | req->assoc_bytes;
	req->info->cmd[cmd_words++] = 0xa0800000 | req->assoc_bytes;
	req->info->cmd[cmd_words++] = 0x25000010;
	req->info->cmd[cmd_words++] = 0x0f020000 | text_bytes;

	switch (req->dir) {
	case EIP97_ENCRYPT:
		req->info->cmd[cmd_words++] = 0x21e60000 | req->auth_bytes;
		break;
	case EIP97_DECRYPT:
		req->info->cmd[cmd_words++] = 0x40e60000 | req->auth_bytes;
		req->info->cmd[cmd_words++] = 0xd0070000 | req->auth_bytes;
		break;
	default:
		return 0;
	}

	return cmd_words;
}

static int eip97_aes_gcm_create_cds(struct eip97_aes_gcm_request *req,
				    u32 cmd_words, u32 *src_nents)
{
	struct eip97_ring *ring = req->ring;
	struct scatterlist *src = req->src;
	volatile struct eip97_cd *cmd;
	int ret;

	*src_nents = 0;

	while (req->src_bytes) {
		if (!src) {
			ret = -EINVAL;
			break;
		} else if (!ring->cmd_prep_num) {
			ret = -EAGAIN;
			break;
		}

		ring->cmd_prep_num--;
		(*src_nents)++;
		cmd = ring->cmd_prep;
		if (++ring->cmd_prep == ring->cmd_base + ring->cmd_num)
			ring->cmd_prep = ring->cmd_base;

		cmd->inbuf = sg_dma_address(src);
		cmd->hdr = 0;

		if (*src_nents == 1) {
			cmd->hdr |= EIP97_DESC_FIRST
				    | EIP97_CD_CT_LEN(cmd_words);
			cmd->ct = req->info_dma;
			cmd->ct_hdr = EIP97_AES_GCM_CT_USE_EIP97_MODE
				      | req->src_bytes;
			cmd->tag = req->tag;
			cmd->tfm = req->info_dma + sizeof(req->info->cmd);
		}

		if (req->src_bytes <= sg_dma_len(src)) {
			cmd->hdr |= EIP97_DESC_BUF_LEN(req->src_bytes);
			req->src_bytes = 0;
			break;
		}

		cmd->hdr |= EIP97_DESC_BUF_LEN(sg_dma_len(src));
		req->src_bytes -= sg_dma_len(src);
		src = sg_next(src);
	}

	if (*src_nents)
		cmd->hdr |= EIP97_DESC_LAST;

	if (req->src_bytes)
		return ret;

	return 0;
}

static int eip97_aes_gcm_create_rds(struct eip97_aes_gcm_request *req,
				    u32 *dst_nents)
{
	struct eip97_ring *ring = req->ring;
	struct scatterlist *dst = req->dst;
	volatile struct eip97_rd *res;
	int ret;

	*dst_nents = 0;

	while (req->dst_bytes) {

		if (!dst) {
			ret = -EINVAL;
			break;
		} else if (!ring->res_prep_num) {
			ret = -EAGAIN;
			break;
		}

		ring->res_prep_num--;
		(*dst_nents)++;
		res = ring->res_prep;
		if (++ring->res_prep == ring->res_base + ring->res_num)
			ring->res_prep = ring->res_base;

		res->outbuf = sg_dma_address(dst);
		res->hdr = 0;

		if (*dst_nents == 1) {
			res->hdr |= EIP97_DESC_FIRST;
			res->tag |= req->tag;
		}

		if (req->dst_bytes <= sg_dma_len(dst)) {
			res->hdr |= EIP97_DESC_BUF_LEN(req->dst_bytes);
			req->dst_bytes = 0;
			break;
		}

		res->hdr |= EIP97_DESC_BUF_LEN(sg_dma_len(dst));
		req->dst_bytes -= sg_dma_len(dst);
		dst = sg_next(dst);
	}

	if (*dst_nents)
		res->hdr |= EIP97_DESC_LAST;

	if (req->dst_bytes)
		return ret;

	return 0;

}

int eip97_aes_gcm_crypt(struct eip97_aes_gcm_request *req)
{
	struct eip97_ring *ring = req->ring;
	volatile struct eip97_rd *res_prep;
	volatile struct eip97_cd *cmd_prep;
	u32 src_nents, dst_nents;
	u32 cmd_prep_num, res_prep_num;
	u32 cmd_words;
	u32 text_bytes;
	int err;

#ifdef EIP97_SAFE_MODE
	//req->key_words is checked always anyway
	if (!req->key)
		return -EINVAL;
	if (!req->prehash)
		return -EINVAL;

	//req->iv_bytes is checked always anyway
	if (!req->iv)
		return -EINVAL;

	switch (req->auth_bytes) {
	case 8:
	case 12:
	case 16:
		break;
	default:
		return -EINVAL;
	}

	if (!req->assoc_bytes)
		return -EINVAL;
	if (!req->src || !req->src_bytes)
		return -EINVAL;
	if (!req->dst || !req->dst_bytes)
		return -EINVAL;
	if (req->dir != EIP97_ENCRYPT && req->dir != EIP97_DECRYPT)
		return -EINVAL;
	if (!req->base)
		return -EINVAL;
	if (req->ring_id < 0 || req->ring_id >= EIP97_RING_MAX)
		return -EINVAL;
	if (!req->ring)
		return -EINVAL;
	if (!req->info)
		return -EINVAL;
	if (!req->info_dma)
		return -EINVAL;
#endif

	switch (req->dir) {
	case EIP97_ENCRYPT:
		text_bytes = req->src_bytes - req->assoc_bytes;
		break;
	case EIP97_DECRYPT:
		text_bytes = req->dst_bytes - req->assoc_bytes;
		break;
	default:
		return -EINVAL;
	}

	err = eip97_aes_gcm_create_tfm(req);
	if (err)
		return err;

	cmd_words = eip97_aes_gcm_create_command_token(req, text_bytes);
	if (!cmd_words)
		return -EINVAL;

	/* Build command descriptors */

	cmd_prep_num = ring->cmd_prep_num;
	cmd_prep = ring->cmd_prep;

	err = eip97_aes_gcm_create_cds(req, cmd_words, &src_nents);
	if (err) {
		ring->cmd_prep_num = cmd_prep_num;
		ring->cmd_prep = cmd_prep;
		return err;
	}

	/* Build result descriptors */

	res_prep_num = ring->res_prep_num;
	res_prep = ring->res_prep;

	err = eip97_aes_gcm_create_rds(req, &dst_nents);
	if (err) {
		ring->cmd_prep_num = cmd_prep_num;
		ring->cmd_prep = cmd_prep;
		ring->res_prep_num = res_prep_num;
		ring->res_prep = res_prep;
		return err;
	}

	/* Sync DMA, if any */
	if (req->sync)
		req->sync(req->sync_ctx);

	/* Make sure all changes to DMA ring are done before we start engine */

	wmb();

	/* Start transfer */

	writel(EIP97_CD_CNT(src_nents),
	       req->base + CDR_PREP_COUNT(req->ring_id));

	writel(EIP97_RD_CNT(dst_nents),
	       req->base + RDR_PREP_COUNT(req->ring_id));

	return -EINPROGRESS;
}
EXPORT_SYMBOL(eip97_aes_gcm_crypt);

void eip97_aead_info_dma_sync(void *ctx)
{
	struct eip97_aead_info_dma_sync *sync_ctx =
		(struct eip97_aead_info_dma_sync *)ctx;

	dma_sync_single_for_device(sync_ctx->dev, sync_ctx->addr,
				   sizeof(struct eip97_aead_info),
				   DMA_TO_DEVICE);
}

static int eip97_aead_enqueue_hard(struct aead_request *req)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);
	struct eip97_aes_gcm_request aes_gcm_req;
	u32 ivsize = crypto_aead_ivsize(crypto_aead_reqtfm(req));
	struct eip97_aead_info_dma_sync sync_ctx;
	unsigned long flags;
	int tag, ret;

	memset(&aes_gcm_req, 0, sizeof(aes_gcm_req));
	memset(&sync_ctx, 0, sizeof(sync_ctx));

	tag = eip97_alloc_tag(tfm_ctx->cryp, &(ctx->type));
	if (tag < 0)
		return tag;

	sync_ctx.dev = tfm_ctx->cryp->dev;
	sync_ctx.addr = ctx->info_dma;

	spin_lock_irqsave(&tfm_ctx->cryp->ring[EIP97_RING0]->lock, flags);

	switch (ctx->mode) {
	case AES_GCM:
		aes_gcm_req.key_words =   tfm_ctx->key_words;
		aes_gcm_req.key =         tfm_ctx->key;
		aes_gcm_req.prehash =     tfm_ctx->prehash;
		aes_gcm_req.iv_bytes =    ivsize;
		aes_gcm_req.iv =          req->iv;
		aes_gcm_req.auth_bytes =  tfm_ctx->auth_bytes;
		aes_gcm_req.assoc_bytes = req->assoclen;
		aes_gcm_req.src =         req->src;
		aes_gcm_req.src_bytes =   ctx->src.proc_bytes;
		aes_gcm_req.dst =         req->dst;
		aes_gcm_req.dst_bytes =   ctx->dst.proc_bytes;
		aes_gcm_req.dir =         ctx->dir;
		aes_gcm_req.tag =         tag;
		aes_gcm_req.base =        tfm_ctx->cryp->base;
		aes_gcm_req.ring_id =     EIP97_RING0;
		aes_gcm_req.ring =        tfm_ctx->cryp->ring[EIP97_RING0];
		aes_gcm_req.info =        &(ctx->info);
		aes_gcm_req.info_dma =    ctx->info_dma;
		aes_gcm_req.sync =        &eip97_aead_info_dma_sync;
		aes_gcm_req.sync_ctx =    &sync_ctx;

		ret = eip97_aes_gcm_crypt(&aes_gcm_req);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	spin_unlock_irqrestore(&tfm_ctx->cryp->ring[EIP97_RING0]->lock, flags);

	return ret;
}

static int eip97_aead_enqueue(struct aead_request *req)
{
	int err;

	err = eip97_aead_enqueue_hard(req);
	if (err == -EINPROGRESS)
		return -EINPROGRESS;

	//TODO: add soft backlog queue in case cd/rd is full!

	return eip97_aead_dma_unmap(req, err, false);
}

static int eip97_aead_dma_map(struct aead_request *req)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);
	struct device *dev = tfm_ctx->cryp->dev;

	ctx->info_dma = dma_map_single(dev, &(ctx->info), sizeof(ctx->info),
				       DMA_TO_DEVICE);

	if (unlikely(dma_mapping_error(dev, ctx->info_dma)))
		goto err_exit;

	if (req->src == req->dst) {
		ctx->src.dma_nents = dma_map_sg(dev, req->src, ctx->src.nents,
						DMA_BIDIRECTIONAL);
		ctx->dst.dma_nents = ctx->src.dma_nents;
		if (unlikely(!ctx->src.dma_nents))
			goto src_map_err;
	} else {
		ctx->src.dma_nents = dma_map_sg(dev, req->src, ctx->src.nents,
						DMA_TO_DEVICE);
		if (unlikely(!ctx->src.dma_nents))
			goto src_map_err;

		ctx->dst.dma_nents = dma_map_sg(dev, req->dst, ctx->dst.nents,
						DMA_FROM_DEVICE);
		if (unlikely(!ctx->dst.dma_nents))
			goto dst_map_err;
	}

	return eip97_aead_enqueue(req);

dst_map_err:
	dma_unmap_sg(dev, req->src, ctx->src.dma_nents, DMA_TO_DEVICE);

src_map_err:
	dma_unmap_single(dev, ctx->info_dma, sizeof(ctx->info), DMA_TO_DEVICE);
err_exit:
	return eip97_aead_dma_unalign(req, -EINVAL, false);
}

static bool eip97_aead_check_aligned(struct scatterlist *sg, size_t *bytes,
				     u32 *nents)
{
	bool ret = true;

	*nents = 0;

	for (*bytes = 0; sg; sg = sg_next(sg)) {

		*bytes += sg->length;

		(*nents)++;

		if (!IS_ALIGNED(sg->offset, sizeof(u32)))
			ret = false;
	}

	return ret;
}

static int eip97_aead_dma_align(struct aead_request *req)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);
	struct scatterlist *aligned_sg;
	bool src_aligned, dst_aligned;
	size_t src_bytes, dst_bytes;
	size_t align_bytes;
	size_t copied;
	u8 *align_buf;

	tfm_ctx->cryp->debug.total_ops++;

	ctx->src.unaligned_sg = req->src;
	ctx->dst.unaligned_sg = req->dst;

	src_aligned = eip97_aead_check_aligned(req->src, &src_bytes,
					       &ctx->src.unaligned_nents);
	if (req->src == req->dst) {
		dst_aligned = src_aligned;
		dst_bytes = src_bytes;
		ctx->dst.unaligned_nents = ctx->src.unaligned_nents;
	} else {
		dst_aligned = eip97_aead_check_aligned(req->dst, &dst_bytes,
						       &ctx->dst.unaligned_nents);
	}

	ctx->src.nents = ctx->src.unaligned_nents;
	ctx->dst.nents = ctx->dst.unaligned_nents;

	if (src_bytes < ctx->src.proc_bytes)
		return eip97_aead_complete(req, -EINVAL, false);
	if (dst_bytes < ctx->dst.proc_bytes)
		return eip97_aead_complete(req, -EINVAL, false);

	if (!src_aligned || !dst_aligned) {
		align_bytes = req->assoclen +
			      req->cryptlen +
			      tfm_ctx->auth_bytes;

		align_buf = kmalloc(align_bytes + sizeof(struct scatterlist),
				    GFP_KERNEL);
		if (!align_buf)
			return eip97_aead_complete(req, -ENOMEM, false);

		aligned_sg = (struct scatterlist *)(align_buf + align_bytes);
		sg_set_buf(aligned_sg, align_buf, align_bytes);

		sg_mark_end(aligned_sg);

		if (!dst_aligned) {
			req->dst = aligned_sg;
			ctx->dst.nents = 1;
		}

		if (!src_aligned) {
			WARN_ONCE(!src_aligned, "EIP97 acceleration of AES-GCM slows down because of unaligned data input!");

			tfm_ctx->cryp->debug.unaligned_ops++;

			copied = sg_copy_to_buffer(req->src, ctx->src.nents,
						   align_buf,
						   ctx->src.proc_bytes);
			req->src = aligned_sg;
			ctx->src.nents = 1;

			if (copied != ctx->src.proc_bytes) {
				kfree(align_buf);
				return eip97_aead_complete(req, -ENODATA,
							   false);
			}
		}
	}

	return eip97_aead_dma_map(req);
}

static int eip97_aes_gcm_enc(struct aead_request *req)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);

	ctx->mode = AES_GCM;
	ctx->dir = EIP97_ENCRYPT;
	ctx->type = EIP97_AEAD;
	ctx->src.proc_bytes = req->assoclen + req->cryptlen;
	ctx->dst.proc_bytes = req->assoclen + req->cryptlen +
			      tfm_ctx->auth_bytes;

	return eip97_aead_dma_align(req);
}

static int eip97_aes_gcm_dec(struct aead_request *req)
{
	struct crypto_tfm *tfm = req->base.tfm;
	struct eip97_aead_tfm_ctx *tfm_ctx = crypto_tfm_ctx(tfm);
	struct eip97_aead_req_ctx *ctx = req_ctx(req);

	ctx->mode = AES_GCM;
	ctx->dir = EIP97_DECRYPT;
	ctx->type = EIP97_AEAD;
	ctx->src.proc_bytes = req->assoclen + req->cryptlen;
	ctx->dst.proc_bytes = req->assoclen + req->cryptlen -
			      tfm_ctx->auth_bytes;

	return eip97_aead_dma_align(req);
}

static int eip97_aead_cra_init(struct crypto_aead *aead)
{
	struct eip97_aead_tfm_ctx *ctx = crypto_tfm_ctx(&aead->base);

	ctx->cryp = eip97_find_dev();
	if (!ctx->cryp) {
		pr_err("can't find crypto device\n");
		return -ENODEV;
	}

	aead->reqsize = sizeof(struct eip97_aead_req_ctx);

	ctx->aes_ctr = crypto_alloc_skcipher("ctr(aes)", 0, CRYPTO_ALG_ASYNC);
	if (IS_ERR(ctx->aes_ctr)) {
		pr_err("Error allocating skcipher 'ctr(aes)'\n");
		return PTR_ERR(ctx->aes_ctr);
	}

	return 0;
}

static void eip97_aead_cra_exit(struct crypto_aead *aead)
{
	struct eip97_aead_tfm_ctx *ctx = crypto_tfm_ctx(&aead->base);

	crypto_free_skcipher(ctx->aes_ctr);
}

struct aead_alg aead_algs[] = {
{
	.setkey		= eip97_aes_gcm_setkey,
	.setauthsize	= eip97_aes_gcm_setauthsize,
	.encrypt	= eip97_aes_gcm_enc,
	.decrypt	= eip97_aes_gcm_dec,
	.ivsize		= 12,
	.init		= eip97_aead_cra_init,
	.exit		= eip97_aead_cra_exit,
	.maxauthsize	= AES_BLOCK_SIZE,

	.base = {
		.cra_name		= "gcm(aes)",
		.cra_driver_name	= "gcm-aes-eip97",
		.cra_priority		= 400,
		.cra_flags		= CRYPTO_ALG_ASYNC,
		.cra_blocksize		= 1,
		.cra_ctxsize		= sizeof(struct eip97_aead_tfm_ctx),
		.cra_alignmask		= 0xf,
		.cra_module		= THIS_MODULE,
	},
},
};


int eip97_aead_register_algs(struct eip97_cryp *cryp)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(aead_algs); i++) {
		err = crypto_register_aead(&aead_algs[i]);
		if (err)
			goto err_aead_algs;

		eip97_test_aead(&aead_algs[i]);
	}

	return 0;

err_aead_algs:
	for (; i--; )
		crypto_unregister_aead(&aead_algs[i]);

	return err;
}

void eip97_aead_unregister_algs(struct eip97_cryp *cryp)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(aead_algs); i++) {
		crypto_unregister_aead(&aead_algs[i]);
	}
}
