/*
 * Driver for EIP97 cryptographic accelerator.
 *
 * Copyright (c) 2018 AVM GmbH <fritzbox_info@avm.de>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This driver is heavily based on the Mediatek driver from Ryder Lee.
 *
 */

#ifndef __EIP97_PLATFORM_H_
#define __EIP97_PLATFORM_H_

#include <linux/interrupt.h>
#include <linux/atomic.h>
#include <linux/workqueue.h>
#include "eip97-regs.h"

/* Transform state control word 0 fields */

#define EIP97_TFM_SHA224		(0x4 << 23)

#define EIP97_TFM_AES_AEAD_DIGEST	(0x2 << 21)

#define EIP97_TFM_AES_128BITS		(0xa << 16)
#define EIP97_TFM_AES_192BITS		(0xc << 16)
#define EIP97_TFM_AES_256BITS		(0xe << 16)

#define EIP97_TFM_CONTAINS_KEY		(0x1 << 16)

#define EIP97_TFM_SIZE(x)		((x) << 8)

#define EIP97_TFM_BASIC_OUT		(0x4 << 0)
#define EIP97_TFM_BASIC_IN		(0x5 << 0)
#define EIP97_TFM_CRYPT_THEN_HASH_OUT	(0x6 << 0)
#define EIP97_TFM_CRYPT_THEN_HASH_IN	(0x7 << 0)
#define EIP97_TFM_HASH_THEN_CRYPT_OUT	(0xe << 0)
#define EIP97_TFM_HASH_THEN_CRYPT_IN	(0xf << 0)

#define EIP97_TFM_ECB			(0x0 << 0)
#define EIP97_TFM_CBC			(0x1 << 0)

/* Transform state control word 1 fields */

#define EIP97_TFM_ENCRYPT_HASH_RESULT	BIT(17)

#define EIP97_TFM_COUNTER_MODE_IV	(0x1 << 10)

#define EIP97_TFM_IV3			BIT(8)
#define EIP97_TFM_IV2			BIT(7)
#define EIP97_TFM_IV1			BIT(6)
#define EIP97_TFM_IV0			BIT(5)

#define EIP97_TFM_CRYPT_MODE_UNDEFINED	(0x2 << 0)
#define EIP97_TFM_CTR_LOAD		(0x6 << 0)



#define SIZE_IN_WORDS(x)	DIV_ROUND_UP((x), 4)

#define SIZE_IN_BYTES(x)	((x) << 2)

enum {
	EIP97_RING0,
	EIP97_RING1,
	EIP97_RING2,
	EIP97_RING3,
	EIP97_RING_MAX
};

#define EIP97_IRQ_NUM		5

#define EIP97_IRQ_RDR0		BIT(1)
#define EIP97_IRQ_RDR1		BIT(3)
#define EIP97_IRQ_RDR2		BIT(5)
#define EIP97_IRQ_RDR3		BIT(7)

#define EIP97_DESC_NUM		512
#define EIP97_DESC_LAST		BIT(22)
#define EIP97_DESC_FIRST	BIT(23)
#define EIP97_DESC_BUF_LEN(x)	(x)
#define EIP97_DESC_TAG(x)	((x) & GENMASK(23, 0))

/**
 * struct eip97_cd - Crypto operation command descriptor
 * @hdr:	descriptor control header
 * @buf:	DMA address of input buffer segment
 * @ct:		DMA address of command token that controls operation flow
 * @ct_hdr:	command token control header
 * @tag:	user-defined field
 * @tfm:	DMA address of transform state
 * @bound:	align descriptors offset boundary
 *
 * Structure passed to the crypto engine to describe where input data
 * needs to be fetched and how it needs to be processed.
 */
struct eip97_cd {
	u32 hdr;
	u32 inbuf;
	u32 ct;
	u32 ct_hdr;
	u32 tag;
	u32 tfm;
	u32 bound[2];
};

#define EIP97_CD_OFF		SIZE_IN_WORDS(sizeof(struct eip97_cd))
#define EIP97_CD_SZ		(EIP97_CD_OFF - 2)
#define EIP97_CD_RING_SZ	((sizeof(struct eip97_cd) * EIP97_DESC_NUM))
#define EIP97_CD_CNT(x)		SIZE_IN_BYTES(EIP97_CD_OFF * (x))
#define EIP97_CD_CT_LEN(x)	((x) << 24)

/**
 * struct eip97_rd - Crypto operation result descriptor
 * @hdr:	descriptor control header
 * @buf:	DMA address of output buffer segment
 * @res1:	result information
 * @res2:	result information
 * @tag:	user-defined field
 * @res3:	result information
 * @bound:	align descriptors offset boundary
 *
 * Structure passed to the crypto engine to describe where output data
 * needs to be pushed to. Also holds additional information on errors or
 * success after the crypto operation finishes.
 */
struct eip97_rd {
	u32 hdr;
	u32 outbuf;
	u32 res1;
	u32 res2;
	u32 tag;
	u32 res3;
	u32 bound[2];
};

#define EIP97_RD_OFF		SIZE_IN_WORDS(sizeof(struct eip97_rd))
#define EIP97_RD_SZ		(EIP97_RD_OFF - 2)
#define EIP97_RD_RING_SZ	((sizeof(struct eip97_rd) * EIP97_DESC_NUM))
#define EIP97_RD_CNT(x)		SIZE_IN_BYTES(EIP97_RD_OFF * (x))
#define EIP97_RD_RESULT_SIZE(x)	((x) & GENMASK(16, 0))
#define EIP97_RD_ERR(x)		(((x) >> 17) & GENMASK(14, 0))
#define EIP97_RD_BUF_OVF	BIT(21)
#define EIP97_RD_DESC_OVF	BIT(20)

typedef u32 eip97_err_t;
#define EIP97_SUCCESS				0
#define EIP97_ERR_PACKET_LENGTH			BIT(0)
#define EIP97_ERR_UNKNOWN_INSTR			BIT(1)
#define EIP97_ERR_TOO_MANY_BYPASS_DATA		BIT(2)
#define EIP97_ERR_CRYPTO_BLOCKSIZE		BIT(3)
#define EIP97_ERR_HASH_BLOCKSIZE		BIT(4)
#define EIP97_ERR_INVALID_MODE_COMBINATION	BIT(5)
#define EIP97_ERR_INVALID_ALGORITHM		BIT(6)
#define EIP97_ERR_HASH_INPUT_OVERFLOW		BIT(7)
#define EIP97_ERR_TTL_HOP_UNDERFLOW		BIT(8)
#define EIP97_ERR_AUTHENTICATION		BIT(9)
#define EIP97_ERR_SN_CHECK			BIT(10)
#define EIP97_ERR_SPI_CHECK			BIT(11)
#define EIP97_ERR_WRONG_CHECKSUM		BIT(12)
#define EIP97_ERR_PAD_VERIFICATION		BIT(13)
#define EIP97_ERR_TIMEOUT			BIT(14)
#define EIP97_ERR_BUFFER_OVERFLOW		BIT(15)
#define EIP97_ERR_DESCRIPTOR_OVERFLOW		BIT(16)

#define EIP97_ERR_CHECK_FAIL	( EIP97_ERR_TTL_HOP_UNDERFLOW \
				| EIP97_ERR_AUTHENTICATION \
				| EIP97_ERR_SN_CHECK \
				| EIP97_ERR_SPI_CHECK \
				| EIP97_ERR_WRONG_CHECKSUM \
				| EIP97_ERR_PAD_VERIFICATION )


/**
 * struct eip97_ring - Ring of command and result descriptors
 *
 * @cmd_base:		pointer to CD ring base
 * @cmd_dma:		DMA address of CD ring
 * @cmd_prep:		pointer to next CD to be prepared for a request
 * @cmd_prep_num:	number of free CDs that can be prepared
 * @cmd_proc:		pointer to next CD to be processed after a request
 * @cmd_proc_num:	number of finished CDs that must be processed
 * @cmd_num:		total number of descriptors in CD ring
 *
 * @res_base:		pointer to RD ring base
 * @res_dma:		DMA address of RD ring
 * @res_prep:		pointer to next RD to be prepared for a request
 * @res_prep_num:	number of free RDs that can be prepared
 * @res_proc:		pointer to next RD to be processed after a request
 * @res_proc_num:	number of finished RDs that must be processed
 * @res_num:		total number of descriptors in RD ring
 *
 * @lock:		locks entire ring except @cmd_proc_num and @res_proc_num
 *
 * Describes a descriptor ring that exchanges crypto operation requests
 * and results with the crypto hardware. It consists of two parts, the
 * command descriptors (CDs) and the result descriptors (RDs).
 * CDs describe the input and RDs the output of a crypto operation.
 */
struct eip97_ring {
	volatile struct eip97_cd *cmd_base;
	dma_addr_t cmd_dma;
	volatile struct eip97_cd *cmd_prep;
	u32 cmd_prep_num;
	volatile struct eip97_cd *cmd_proc;
	atomic_t cmd_proc_num;
	u32 cmd_num;

	volatile struct eip97_rd *res_base;
	dma_addr_t res_dma;
	volatile struct eip97_rd *res_prep;
	u32 res_prep_num;
	volatile struct eip97_rd *res_proc;
	atomic_t res_proc_num;
	u32 res_num;

	spinlock_t lock;
};

enum eip97_op_type {
	EIP97_INVALID_OP_TYPE = 0,
	EIP97_ABLKCIPHER,
	EIP97_AEAD,
	EIP97_AHASH
};

enum crypt_direction {
	EIP97_ENCRYPT,
	EIP97_DECRYPT
};

struct eip97_debug {
	u32 unaligned_ops;
	u32 total_ops;

	struct dentry *dir;
};

/**
 * struct eip97_cryp - Cryptographic device
 * @base:		pointer to mapped register I/O base
 * @irq:		global system and rings IRQ
 * @dev:		pointer to device
 * @clk_cryp:		pointer to crypto clock
 * @ring:		array of ring descriptions
 * @irq_bh:		tasklet for IRQ postprocessing
 * @worker:		workqueue for IRQ postpostprocessing
 * @tags:		tag lookup table for requests on ring 0,
 *			with at least one tag per possible request
 * @taglock:		spinlock to protect @tags
 * @dev_list:		list of EIP97 devices
 * @debug:		debug statistics
 *
 * Structure storing cryptographic device information.
 */
struct eip97_cryp {
	void __iomem *base;
	int irq[EIP97_IRQ_NUM];
	struct device *dev;
	struct clk *clk_cryp;

	struct eip97_ring *ring[EIP97_RING_MAX];

	struct tasklet_struct irq_bh;

	enum eip97_op_type *tags[EIP97_DESC_NUM];
	spinlock_t taglock;

	struct list_head dev_list;

	struct eip97_debug debug;
};

struct eip97_cryp *eip97_find_dev(void);
int eip97_alloc_tag(struct eip97_cryp *cryp, enum eip97_op_type *op_type);

static inline void eip97_write_state_le(__le32 *dst, const u32 *src, u32 words)
{
	int i;

	for (i = 0; i < words; i++)
		dst[i] = cpu_to_le32(src[i]);
}

static inline void eip97_write_state_be(__be32 *dst, const u32 *src, u32 words)
{
	int i;

	for (i = 0; i < words; i++)
		dst[i] = cpu_to_be32(src[i]);
}

#endif /* __EIP97_PLATFORM_H_ */
