/*
 * Driver for EIP97 cryptographic accelerator.
 *
 * Copyright (c) 2018 AVM GmbH <fritzbox_info@avm.de>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This driver is heavily based on the Mediatek driver from Ryder Lee.
 *
 */

#include <crypto/internal/aead.h>
#include <crypto/internal/hash.h>
#include <crypto/algapi.h>
#include <linux/scatterlist.h>
#include "eip97-utest.h"

//TODO: Testfall mit identischer src/dst-Scatterlist!
struct test_template tests[] = {
{
	.cra_name = "gcm(aes)",
	.attrs.aead = {
		.key = {
			.data = "\xb2\xb6\x3c\xac\xb2\x4a\x27\x63"
				"\x7f\xd5\xca\x03\xd5\x28\x6c\x0e",
			.len  = 16,
		},
		.out = {
			.data      = "\x9c\x8d\xdb\xf4\xe6\x7d\xaa\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\xee\x36\x45\xee\x6d\x70\x95\x88"
				     "\xd1\x79\x3a\xee\xf0\x58\x09\x12",
			.len       = 48,
			.chunks    = { 16, 39 },
			.misaligns = { 3, 0 },
			.nchunks   = 2,
		},
		.in = {
			.data      = "\x9c\x8d\xdb\xf4\xe6\x7d\xaa\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x05\xa3\xba\x6a\xc4\xa4\x03\xdc"
				     "\x44\x45\x85\xb9\x1c\xff\xa9\x79"
				     "\x9f\xfc\x17\xe2\xa6\x93\x8e\xbe"
				     "\xf5\xd8\x01\xe0\x5f\xc1\x12\x7b"
				     "\x70\x52\x0c\x78\xc9\x16\xce\x48"
				     "\xfd\x63\xf0\x26\x38\x6e\xe6\x7d",
			.len       = 64,
			.chunks    = { 16, 32, 18 },
			.misaligns = { 1, 0, 2 },
			.nchunks   = 3,
		},
		.iv = {
			.data = "\x4c\xcf\x86\x7a\xe9\xd0\x57\x91"
				"\xdc\x3d\xc0\x08\x2a\x40\xe7\x85",
			.len  = 16,
		},
		.textlen = 32,
		.assoclen = 16,
		.taglen = 16,
		.dir = DECRYPT,
	},
},
/*
{
	.cra_name = "gcm(aes)",
	.attrs.aead = {
		.key = {
			.data = "\xb2\xb6\x3c\xac\xb2\x4a\x27\x63"
				"\x7f\xd5\xca\x03\xd5\x28\x6c\x0e",
			.len  = 16,
		},
		.in = {
			.data      = "\x9c\x8d\xdb\xf4\xe6\x7d\xaa\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\xee\x36\x45\xee\x6d\x70\x95\x88"
				     "\xd1\x79\x3a\xee\xf0\x58\x09\x12",
			.len       = 48,
			.chunks    = { 16, 39 },
			.misaligns = { 3, 0 },
			.nchunks   = 2,
		},
		.out = {
			.data      = "\x9c\x8d\xdb\xf4\xe6\x7d\xaa\xdf"
				     "\x0e\xde\x7d\xb2\x43\x36\x11\xdf"
				     "\x05\xa3\xba\x6a\xc4\xa4\x03\xdc"
				     "\x44\x45\x85\xb9\x1c\xff\xa9\x79"
				     "\x9f\xfc\x17\xe2\xa6\x93\x8e\xbe"
				     "\xf5\xd8\x01\xe0\x5f\xc1\x12\x7b"
				     "\x70\x52\x0c\x78\xc9\x16\xce\x48"
				     "\xfd\x63\xf0\x26\x38\x6e\xe6\x7d",
			.len       = 64,
			.chunks    = { 16, 32, 18 },
			.misaligns = { 1, 0, 2 },
			.nchunks   = 3,
		},
		.iv = {
			.data = "\x4c\xcf\x86\x7a\xe9\xd0\x57\x91"
				"\xdc\x3d\xc0\x08\x2a\x40\xe7\x85",
			.len  = 16,
		},
		.textlen = 32,
		.assoclen = 16,
		.taglen = 16,
		.dir = ENCRYPT,
	},
},
*/
};

struct callback_context {
	void *alg;
	void *test_data;
	u8 *checkbuf;
};

static struct test_template *eip97_find_test(const char *cra_name)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(tests); i++) {
		if (!strcmp(cra_name, tests[i].cra_name)) {
			return &tests[i];
		}
	}

	return ERR_PTR(-EINVAL);
}

static struct scatterlist *eip97_alloc_pile(struct template_attr *attr,
					    bool copy)
{
	struct scatterlist *sgs;
	u32 misalign, chunk_len;
	u32 left = attr->len;
	u32 virtual_len;
	u32 from = 0;
	int i, err;
	u8 *chunk;

	if (attr->nchunks < 1)
		return ERR_PTR(-EINVAL);

	sgs = kcalloc(attr->nchunks, sizeof(struct scatterlist), GFP_KERNEL);
	if (!sgs)
		return ERR_PTR(-ENOMEM);

	for (i = 0; i < attr->nchunks && i < MAX_CHUNKS; i++) {
		misalign  = attr->misaligns[i];
		chunk_len = attr->chunks[i];

		//kmalloc() result guaranteed to be word-aligned
		chunk = kmalloc(misalign + chunk_len, GFP_KERNEL);
		if (!chunk) {
			err = -ENOMEM;
			goto memory_error;
		}

		if (chunk_len < left) {
			virtual_len = chunk_len;
		} else {
			virtual_len = left;
		}

		if (copy)
			memcpy(chunk + misalign, attr->data + from, virtual_len);

		//length is chunk_len instead of virtual_len
		//to confuse driver if chunk_len > left
		sg_set_buf(&sgs[i], chunk + misalign, chunk_len);

		from += virtual_len;
		left -= virtual_len;
	}

	if (left > 0) {
		err = -EINVAL;
		goto chunk_error;
	}

	sg_mark_end(&sgs[attr->nchunks - 1]);

	return sgs;

chunk_error:
memory_error:
	while (i--) {
		kfree(((u8 *)sg_virt(&sgs[i])) - attr->misaligns[i]);
	}
	kfree(sgs);

	return ERR_PTR(err);
}

static void eip97_free_pile(struct template_attr *attr, struct scatterlist *sg)
{
	struct scatterlist *sgs = sg;
	int i = 0;

	for (; sg; sg = sg_next(sg))
		kfree(((u8 *)sg_virt(sg)) - attr->misaligns[i++]);

	kfree(sgs);
}

static void eip97_aead_done(struct crypto_async_request *areq, int err)
{
	struct crypto_aead *tfm = __crypto_aead_cast(areq->tfm);
	struct aead_request *req = aead_request_cast(areq);
	struct callback_context *ctx = req->base.data;
	struct aead_template *test_data =
				(struct aead_template *)ctx->test_data;
	struct aead_alg *alg = (struct aead_alg *)ctx->alg;
	size_t copied, cmplen;
	int inequal;

	if (err) {
		pr_err("EIP97 accelerated crypto algorithm '%s' failed self-check with error %d.\n",
		       alg->base.cra_driver_name, err);
		goto cleanup;
	}

	copied = sg_copy_to_buffer(req->dst, sg_nents(req->dst),
				   ctx->checkbuf, test_data->out.len);

	if (test_data->dir == ENCRYPT)
		cmplen = test_data->textlen + test_data->taglen;
	else
		cmplen = test_data->textlen;

	inequal = memcmp(ctx->checkbuf       + test_data->assoclen,
			 test_data->out.data + test_data->assoclen,
			 cmplen);

	if (copied != test_data->out.len || inequal) {
		pr_err("EIP97 accelerated crypto algorithm '%s' failed self-check due to wrong output.\n",
		       alg->base.cra_driver_name);
		goto cleanup;
	}

	if (alg)
		alg->base.cra_flags |= CRYPTO_ALG_TESTED;
cleanup:
	eip97_free_pile(&(test_data->in), req->src);
	eip97_free_pile(&(test_data->out), req->dst);

	kfree(req);
	kfree(ctx);

	crypto_free_aead(tfm);
}

void eip97_aead_request(struct crypto_aead *tfm,
			struct aead_template *test_data,
			struct aead_alg *alg)
{
	struct scatterlist *src, *dst;
	struct callback_context *ctx;
	struct aead_request *req;
	unsigned int cryptlen;
	int err;

	err = crypto_aead_setkey(tfm, test_data->key.data,
				 test_data->key.len);
	if (err)
		goto error_setkey;

	err = crypto_aead_setauthsize(tfm, test_data->taglen);
	if (err)
		goto error_setauthsize;

	src = eip97_alloc_pile(&test_data->in, true);
	if (IS_ERR(src))
		goto error_pile_in;

	dst = eip97_alloc_pile(&test_data->out, false);
	if (IS_ERR(dst))
		goto error_pile_out;

	req = aead_request_alloc(tfm, GFP_KERNEL);
	if (!req)
		goto error_request;

	aead_request_set_ad(req, test_data->assoclen);

	if (test_data->dir == ENCRYPT)
		cryptlen = test_data->textlen;
	else
		cryptlen = test_data->textlen + test_data->taglen;

	aead_request_set_crypt(req, src, dst, cryptlen,
			       (u8 *)test_data->iv.data);

	ctx = kmalloc(sizeof(*ctx) + test_data->out.len, GFP_KERNEL);
	if (!ctx)
		goto error_context;

	ctx->alg = alg;
	ctx->test_data = test_data;
	ctx->checkbuf = (u8 *)(ctx+1);

	aead_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				  eip97_aead_done, ctx);

	switch (test_data->dir) {
	case ENCRYPT:
		err = crypto_aead_encrypt(req);
		break;
	case DECRYPT:
		err = crypto_aead_decrypt(req);
		break;
	default:
		err = -EINVAL;
	}

	return; //on error, eip97_aead_done freed memory

error_context:
	aead_request_free(req);

error_request:
	eip97_free_pile(&test_data->out, dst);

error_pile_out:
	eip97_free_pile(&test_data->in, src);

error_pile_in:
error_setauthsize:
error_setkey:
	crypto_free_aead(tfm);
}

void eip97_test_aead(struct aead_alg *alg)
{
	struct test_template *test_data;
	struct crypto_aead *tfm;

	alg->base.cra_flags &= (~CRYPTO_ALG_TESTED);

	test_data = eip97_find_test(alg->base.cra_name);
	if (IS_ERR(test_data))
		return;

	/*
	 * Temporarily declare algorithm as tested so crypto_alloc_aead()
	 * finds it. this is legit as the tested flag can also be set if there
	 * was no test, which until now is true.
	 */
	alg->base.cra_flags |= CRYPTO_ALG_TESTED;

	tfm = crypto_alloc_aead(alg->base.cra_driver_name, 0, 0);

	alg->base.cra_flags &= (~CRYPTO_ALG_TESTED);

	if (IS_ERR(tfm))
		return;

	eip97_aead_request(tfm, &test_data->attrs.aead, alg);
}
